################################################################################
# This is the Sage equivalent of the Java code in Example 2.23 in Section 2.6. #
# It uses Newton's algorithm to find the critical points of the function       #
# f(x,y) = x^3 - x*y - x + x*y^3 - y^4, starting with the initial point (0,0). #
# To use other initial points just replace the coordinates (0,0) when running  #
# the newton function at the end, e.g. newton(-1-1) or newton(-5,-5) instead   #
# of newton(0,0).                                                              #
################################################################################
var('x,y')
f(x,y) = x^3 - x*y - x + x*y^3 - y^4
fx(x,y) = f.derivative(x)       # First partial derivative of f wrt x
fy(x,y) = f.derivative(y)       # First partial derivative of f wrt y
fxx(x,y) = f.derivative(x,2)    # Second partial derivative of f wrt x
fyy(x,y) = f.derivative(y,2)    # Second partial derivative of f wrt y
fxy(x,y) = f.derivative(x,y)    # Mixed second partial derivative of f wrt x & y
def newton(x,y):
    x = float(x)    # Convert to floating-point numbers to avoid problems
    y = float(y)
    for n in xrange(1,101):
        D = fxx(x,y)*fyy(x,y) - fxy(x,y)^2
        xn = x
        yn = y
        if D == 0:
            print "Error: D = 0 at iteration n = %d" % (n)
            exit
        else:
            x = xn - (fyy(xn,yn)*fx(xn,yn) - fxy(xn,yn)*fy(xn,yn))/D
            y = yn - (fxx(xn,yn)*fy(xn,yn) - fxy(xn,yn)*fx(xn,yn))/D
            print "n = %d: (%.15f,%.15f)" % (n,x,y)
# Now use the initial point (0,0)
newton(0,0)
